@@ -68,7 +68,7 @@ vec_proxy_equal.POSIXlt <- function(x, ...) {
 #' @export
 vec_proxy_compare.POSIXlt <- function(x, ...) {
   x <- vec_cast(x, new_datetime(tzone = tzone(x)))
-  vec_proxy_compare(x, ...)
+  vec_proxy_compare(x)
 }
 
 #' @export
