@@ -84,7 +84,7 @@ SEXP vec_as_custom_names(SEXP names, const struct name_repair_opts* opts) {
 
   // Don't use vctrs dispatch utils because we match argument positionally
   SEXP call = PROTECT(Rf_lang2(syms_repair, syms_names));
-  SEXP mask = PROTECT(r_new_environment(R_GlobalEnv, 2));
+  SEXP mask = PROTECT(r_new_environment(R_GlobalEnv));
   Rf_defineVar(syms_repair, opts->fn, mask);
   Rf_defineVar(syms_names, names, mask);
   SEXP out = PROTECT(Rf_eval(call, mask));
@@ -95,10 +95,11 @@ SEXP vec_as_custom_names(SEXP names, const struct name_repair_opts* opts) {
   return out;
 }
 
+static
+SEXP vec_names_impl(SEXP x, bool proxy) {
+  bool has_class = OBJECT(x);
 
-// [[ register(); include("vctrs.h") ]]
-SEXP vec_names(SEXP x) {
-  if (OBJECT(x) && Rf_inherits(x, "data.frame")) {
+  if (has_class && Rf_inherits(x, "data.frame")) {
     // Only return row names if they are character. Data frames with
     // automatic row names are treated as unnamed.
     SEXP rn = df_rownames(x);
@@ -109,15 +110,15 @@ SEXP vec_names(SEXP x) {
     }
   }
 
-  if (vec_dim_n(x) == 1) {
-    if (OBJECT(x)) {
+  if (vec_bare_dim(x) == R_NilValue) {
+    if (!proxy && has_class) {
       return vctrs_dispatch1(syms_names, fns_names, syms_x, x);
     } else {
       return r_names(x);
     }
   }
 
-  SEXP dimnames = PROTECT(Rf_getAttrib(x, R_DimNamesSymbol));
+  SEXP dimnames = PROTECT(r_attrib_get(x, R_DimNamesSymbol));
   if (dimnames == R_NilValue || Rf_length(dimnames) < 1) {
     UNPROTECT(1);
     return R_NilValue;
@@ -128,6 +129,15 @@ SEXP vec_names(SEXP x) {
   return out;
 }
 
+// [[ register(); include("vctrs.h") ]]
+SEXP vec_names(SEXP x) {
+  return vec_names_impl(x, false);
+}
+// [[ include("vctrs.h") ]]
+SEXP vec_proxy_names(SEXP x) {
+  return vec_names_impl(x, true);
+}
+
 // [[ register() ]]
 SEXP vctrs_as_minimal_names(SEXP names) {
   if (TYPEOF(names) != STRSXP) {
@@ -527,7 +537,11 @@ SEXP apply_name_spec(SEXP name_spec, SEXP outer, SEXP inner, R_len_t n) {
   }
 
   if (outer == strings_empty || outer == NA_STRING) {
-    return inner;
+    if (inner == R_NilValue) {
+      return chrs_empty;
+    } else {
+      return inner;
+    }
   }
 
   if (r_is_empty_names(inner)) {
@@ -556,22 +570,23 @@ SEXP apply_name_spec(SEXP name_spec, SEXP outer, SEXP inner, R_len_t n) {
   }
   PROTECT(name_spec);
 
-  // Recycle `outer` so specs don't need to refer to both `outer` and `inner`
-  SEXP outer_chr = PROTECT(Rf_allocVector(STRSXP, n));
-  r_chr_fill(outer_chr, outer, n);
+  SEXP outer_chr = PROTECT(r_str_as_character(outer));
 
-  SEXP out = vctrs_dispatch2(syms_dot_name_spec, name_spec,
-                             syms_outer, outer_chr,
-                             syms_inner, inner);
+  SEXP out = PROTECT(vctrs_dispatch2(syms_dot_name_spec, name_spec,
+                                     syms_outer, outer_chr,
+                                     syms_inner, inner));
+  out = vec_recycle(out, n, NULL);
 
-  if (TYPEOF(out) != STRSXP) {
-    Rf_errorcall(R_NilValue, "`.name_spec` must return a character vector.");
-  }
-  if (Rf_length(out) != n) {
-    Rf_errorcall(R_NilValue, "`.name_spec` must return a character vector as long as `inner`.");
+  if (out != R_NilValue) {
+    if (TYPEOF(out) != STRSXP) {
+      Rf_errorcall(R_NilValue, "`.name_spec` must return a character vector.");
+    }
+    if (Rf_length(out) != n) {
+      Rf_errorcall(R_NilValue, "`.name_spec` must return a character vector as long as `inner`.");
+    }
   }
 
-  UNPROTECT(3);
+  UNPROTECT(4);
   return out;
 }
 
@@ -684,8 +699,8 @@ static void check_names(SEXP x, SEXP names) {
   }
 }
 
-SEXP vec_set_rownames(SEXP x, SEXP names) {
-  if (OBJECT(x)) {
+SEXP vec_set_rownames(SEXP x, SEXP names, bool proxy, const enum vctrs_owned owned) {
+  if (!proxy && OBJECT(x)) {
     return set_rownames_fallback(x, names);
   }
 
@@ -700,10 +715,13 @@ SEXP vec_set_rownames(SEXP x, SEXP names) {
     }
   }
 
-  x = PROTECT_N(r_clone_referenced(x), &nprot);
+  x = PROTECT_N(vec_clone_referenced(x, owned), &nprot);
 
   if (dim_names == R_NilValue) {
     dim_names = PROTECT_N(Rf_allocVector(VECSXP, vec_dim_n(x)), &nprot);
+  } else {
+    // Also clone attribute
+    dim_names = PROTECT_N(Rf_shallow_duplicate(dim_names), &nprot);
   }
 
   SET_VECTOR_ELT(dim_names, 0, names);
@@ -714,13 +732,13 @@ SEXP vec_set_rownames(SEXP x, SEXP names) {
   return x;
 }
 
-SEXP vec_set_df_rownames(SEXP x, SEXP names) {
+SEXP vec_set_df_rownames(SEXP x, SEXP names, bool proxy, const enum vctrs_owned owned) {
   if (names == R_NilValue) {
     if (rownames_type(df_rownames(x)) != ROWNAMES_IDENTIFIERS) {
       return(x);
     }
 
-    x = PROTECT(r_clone_referenced(x));
+    x = PROTECT(vec_clone_referenced(x, owned));
     init_compact_rownames(x, vec_size(x));
 
     UNPROTECT(1);
@@ -728,9 +746,12 @@ SEXP vec_set_df_rownames(SEXP x, SEXP names) {
   }
 
   // Repair row names silently
-  names = PROTECT(vec_as_names(names, p_unique_repair_silent_opts));
+  if (!proxy) {
+    names = vec_as_names(names, p_unique_repair_silent_opts);
+  }
+  PROTECT(names);
 
-  x = PROTECT(r_clone_referenced(x));
+  x = PROTECT(vec_clone_referenced(x, owned));
   Rf_setAttrib(x, R_RowNamesSymbol, names);
 
   UNPROTECT(2);
@@ -739,19 +760,18 @@ SEXP vec_set_df_rownames(SEXP x, SEXP names) {
 
 // FIXME: Do we need to get the vec_proxy() and only fall back if it doesn't
 // exist? See #526 and #531 for discussion and the related issue.
-// [[ include("utils.h"); register() ]]
-SEXP vec_set_names(SEXP x, SEXP names) {
+SEXP vec_set_names_impl(SEXP x, SEXP names, bool proxy, const enum vctrs_owned owned) {
   check_names(x, names);
 
   if (is_data_frame(x)) {
-    return vec_set_df_rownames(x, names);
+    return vec_set_df_rownames(x, names, proxy, owned);
   }
 
   if (has_dim(x)) {
-    return vec_set_rownames(x, names);
+    return vec_set_rownames(x, names, proxy, owned);
   }
 
-  if (OBJECT(x)) {
+  if (!proxy && OBJECT(x)) {
     return set_names_fallback(x, names);
   }
 
@@ -760,12 +780,21 @@ SEXP vec_set_names(SEXP x, SEXP names) {
     return x;
   }
 
-  x = PROTECT(r_clone_referenced(x));
+  x = PROTECT(vec_clone_referenced(x, owned));
   Rf_setAttrib(x, R_NamesSymbol, names);
 
   UNPROTECT(1);
   return x;
 }
+// [[ include("utils.h"); register() ]]
+SEXP vec_set_names(SEXP x, SEXP names) {
+  return vec_set_names_impl(x, names, false, VCTRS_OWNED_false);
+}
+// [[ include("utils.h") ]]
+SEXP vec_proxy_set_names(SEXP x, SEXP names, const enum vctrs_owned owned) {
+  return vec_set_names_impl(x, names, true, owned);
+}
+
 
 SEXP vctrs_validate_name_repair_arg(SEXP arg) {
   struct name_repair_opts opts = new_name_repair_opts(arg, args_empty, true);
@@ -809,7 +838,7 @@ struct name_repair_opts new_name_repair_opts(SEXP name_repair, struct vctrs_arg*
     } else if (c == strings_check_unique) {
       opts.type = name_repair_check_unique;
     } else {
-      Rf_errorcall(R_NilValue, "`.name_repair` can't be \"%s\". See `?vctrs::vec_as_names`.", CHAR(name_repair));
+      Rf_errorcall(R_NilValue, "`.name_repair` can't be \"%s\". See `?vctrs::vec_as_names`.", CHAR(c));
     }
 
     return opts;
