@@ -1,4 +1,3 @@
-context("test-type-data-frame")
 
 # printing ----------------------------------------------------------------
 
@@ -94,7 +93,7 @@ test_that("combining data frames with foreign classes uses fallback", {
 
   # There should be only one warning even if many fallbacks
   expect_length(cnds, 1)
-  expect_is(cnds[[1]], "warning")
+  expect_s3_class(cnds[[1]], "warning")
   expect_match(cnds[[1]]$message, "falling back to <data.frame>")
 
   expect_incompatible_df(
@@ -110,8 +109,9 @@ test_that("combining data frames with foreign classes uses fallback", {
     foo <- structure(mtcars[1:3], class = c("foo", "data.frame"))
     bar <- structure(mtcars[4:6], class = c("bar", "data.frame"))
     baz <- structure(mtcars[7:9], class = c("baz", "data.frame"))
-    expect_warning(vec_ptype_common_df_fallback(foo, bar, baz))
-    expect_warning(vec_ptype_common_df_fallback(foo, baz, bar, baz, foo, bar))
+    # Nested expect_warning() require testthat > 2.3.2
+    suppressWarnings(expect_warning(vec_ptype_common_df_fallback(foo, bar, baz)))
+    suppressWarnings(expect_warning(vec_ptype_common_df_fallback(foo, baz, bar, baz, foo, bar)))
 
     with_fallback_warning(expect_df_fallback_warning(invisible(vec_rbind(foo, data.frame(), foo))))
     with_fallback_warning(expect_df_fallback_warning(invisible(vec_cbind(foo, data.frame(x = 1)))))
@@ -220,8 +220,8 @@ test_that("can restore subclasses of data frames", {
 })
 
 test_that("df_cast() checks for names", {
-  x <- data_frame(1)
-  y <- data_frame(2)
+  x <- new_data_frame(list(1))
+  y <- new_data_frame(list(2))
   expect_error(vec_cast_common(x, y), "must have names")
 })
 
@@ -409,6 +409,100 @@ test_that("new_data_frame() zaps existing attributes", {
   )
 })
 
+# data_frame --------------------------------------------------------------
+
+test_that("can construct data frames with empty input", {
+  expect_identical(data_frame(), new_data_frame())
+  expect_named(data_frame(), character())
+})
+
+test_that("input is tidy recycled", {
+  expect_identical(
+    data_frame(x = 1, y = 1:3),
+    data_frame(x = c(1, 1, 1), y = 1:3)
+  )
+
+  expect_identical(
+    data_frame(x = 1, y = integer()),
+    data_frame(x = double(), y = integer())
+  )
+
+  expect_error(data_frame(1:2, 1:3), class = "vctrs_error_incompatible_size")
+})
+
+test_that("dots are dynamic", {
+  list_2_data_frame <- function(x) data_frame(!!!x)
+
+  expect_identical(
+    list_2_data_frame(list(x = 1, y = 2)),
+    data_frame(x = 1, y = 2)
+  )
+})
+
+test_that("unnamed input is auto named with empty strings", {
+  expect_named(data_frame(1, 2, .name_repair = "minimal"), c("", ""))
+})
+
+test_that("unnamed data frames are auto spliced", {
+  expect_identical(
+    data_frame(w = 1, data_frame(x = 2, y = 3), z = 4),
+    data_frame(w = 1, x = 2, y = 3, z = 4)
+  )
+})
+
+test_that("named data frames are not spliced", {
+  df_col <- data_frame(x = 2, y = 3)
+  df <- data_frame(w = 1, col = data_frame(x = 2, y = 3), z = 4)
+
+  expect_identical(df$col, df_col)
+})
+
+test_that("spliced data frames without names are caught", {
+  df_col <- new_data_frame(list(1))
+  expect_error(data_frame(df_col), "corrupt data frame")
+})
+
+test_that("`NULL` inputs are dropped", {
+  expect_identical(data_frame(NULL, x = 1, NULL), data_frame(x = 1))
+})
+
+test_that("`NULL` inputs are dropped before name repair", {
+  expect_identical(
+    data_frame(x = NULL, x = 1, .name_repair = "check_unique"),
+    data_frame(x = 1)
+  )
+})
+
+test_that("`.size` can force a desired size", {
+  df <- data_frame(x = 1, .size = 5)
+  expect_identical(df$x, rep(1, 5))
+
+  expect_size(data_frame(.size = 5), 5L)
+})
+
+test_that("`.name_repair` repairs names", {
+  expect_named(
+    expect_message(data_frame(x = 1, x = 1, .name_repair = "unique")),
+    c("x...1", "x...2")
+  )
+})
+
+test_that("`.name_repair` happens after auto-naming with empty strings", {
+  expect_named(
+    expect_message(data_frame(1, 2, .name_repair = "unique")),
+    c("...1", "...2")
+  )
+})
+
+test_that("`.name_repair` happens after splicing", {
+  expect_named(
+    expect_message(data_frame(x = 1, data_frame(x = 2), .name_repair = "unique")),
+    c("x...1", "x...2")
+  )
+})
+
+# fallback ----------------------------------------------------------------
+
 test_that("data frame fallback handles column types (#999)", {
   df1 <- foobar(data.frame(x = 1))
   df2 <- foobar(data.frame(x = 1, y = 2))
@@ -466,9 +560,9 @@ test_that("data frame fallback handles column types (#999)", {
 
 test_that("falls back to tibble for tibble subclasses (#1025)", {
   foo <- foobar(tibble::as_tibble(mtcars))
-  expect_is(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars)), "tbl_df")
-  expect_is(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, mtcars)), "tbl_df")
-  expect_is(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, foobar(mtcars))), "tbl_df")
+  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars)), "tbl_df")
+  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, mtcars)), "tbl_df")
+  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, foobar(mtcars))), "tbl_df")
 
   verify_errors({
     with_fallback_warning(expect_df_fallback_warning(
